#include "common/hashtable.h"

// Non-cryptographic hashing algorithm optimized for strings
//
// Since characters are usually between 0x30 and 0x7A, operations are done
// to evenly distribute the hashes, assuming that all letters have the same
// occurence.
//
// After these operations, a simple character folding algorithm is
// applied, based on multiplying by a prime number, choosen here for
// operations on chars.
unsigned char common_hashtable_compute_hash(const char *str, unsigned char modulus) {
	unsigned char result = 0;
	unsigned char extended_char;

	while(*str != '\0') {
		extended_char = ((unsigned char) *str - 0x3A);
		extended_char += (extended_char & 0x01) << 7;
		
		result += extended_char + COMMON_HASHTABLE_PRIME * result;
		str++;
	}

	return result % modulus;
}

// Allocates the hashtable content
//
// Open addressing tables allows to pre-initialize all elements;
// this task is handled on initialization, and tables allocated by this mean
// should be freed later with `common_hashtable_free`.
bool common_hashtable_initialize(struct CommonHashtableTable *hashtable, unsigned char hashtable_size) {
	hashtable->base = (struct CommonHashtableElement *) calloc(hashtable_size, sizeof(struct CommonHashtableElement));
	hashtable->size = hashtable_size;

	if(!hashtable->base) {
		return false;
	}

	return true;
}

// Free the hashtable content
//
// Handles freeing the table initialized by `common_hashtable_initialize`
// as well as strings allocated on inserting elements.
void common_hashtable_free(struct CommonHashtableTable *hashtable) {
	unsigned char i;
	struct CommonHashtableElement *el;

	// Free allocated names
	for(i = 0; i < hashtable->size; i++) {
		el = hashtable->base + i;
		if(el->name != NULL) free((void *) el->name);
	}

	free((void *) hashtable->base);
}

// Regular hashtable find algorithm for linear open addressing tables
//
// Elements are stored in a plane of the hashtable size, and collisions are
// solved by finding the first empty place after the theoretical (hash-given)
// element's position.
struct CommonHashtableElement *common_hashtable_find(struct CommonHashtableTable *hashtable, const char *name) {
	struct CommonHashtableElement *el;
	unsigned char initial_pos;
	unsigned char pos;

	initial_pos = common_hashtable_compute_hash(name, hashtable->size);
	pos = initial_pos;

	while(el = hashtable->base + pos, el->name != NULL) {
		if(strcmp(el->name, name) == 0) return el;
		if(++pos >= hashtable->size) pos = 0;
		// Ensure to fail when the table is full
		if(pos == initial_pos) return false;
	}

	return NULL;
}

bool common_hashtable_insert(struct CommonHashtableTable *hashtable, const char *name, void *content) {
	struct CommonHashtableElement *el;
	unsigned char initial_pos;
	unsigned char pos;

	initial_pos = common_hashtable_compute_hash(name, hashtable->size);
	pos = initial_pos;

	while(el = hashtable->base + pos, el->name != NULL) {
		// Check if an element with the same name already exists
		if(strcmp(el->name, name) == 0) return false;
		if(++pos >= hashtable->size) pos = 0;
		// Ensure to fail when the table is full
		if(pos == initial_pos) return false;
	}

	// Clone name into a heap-allocated string
	el->name = (char *) calloc(COMMON_HASHTABLE_KEY_LENGTH + 1, sizeof(char));
	if(!el->name) {
		return false;
	}

	strncpy(el->name, name, COMMON_HASHTABLE_KEY_LENGTH);
	el->content = content;

	return true;
}

