#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "common/hashtable.h"
#include "libcapulet/net_udp.h"
#include "libcapulet/rdma_ib.h"

#define RDMA_TEST_FILE 1
#define RDMA_TEST_STRESS 2

struct ServeMrThreadContext {
	struct CommonHashtableTable *mr_mgr;
	int server_socket;
};

void *serve_mr_td_fn(void *raw_ctx) {
	struct ServeMrThreadContext *ctx = (struct ServeMrThreadContext *) raw_ctx;
	capulet_net_udp_serve_mr(ctx->mr_mgr, ctx->server_socket);
	return raw_ctx;
}

int main(int argc, char *argv[]) {
	struct CapuletRdmaIbContext rdma_ctx;
	struct CapuletNetUdpContext *udp_ctx;
	struct CapuletNetUdpMrInfoPacket in_info_packet;
	struct CapuletNetUdpMrInfoPacket out_info_packet;
	struct CommonHashtableTable mr_mgr;

	struct ServeMrThreadContext serve_mr_td_ctx;

	struct ibv_mr *in_mr;
	struct ibv_mr *out_mr;
	struct ibv_wc poll_wc;

	struct addrinfo server_hints;
	struct addrinfo *server_infos;
	struct timespec timestamps[2];
	pthread_t serve_mr_td;

	ssize_t read_count;
	size_t remote_host_size;
	size_t device_name_size;
	long elapsed_time;
	long max_time;
	int allocated_size;
	int server_socket;
	int poll_result;
	int opt;
	char *remote_host;
	char *device_name;
	char in_mr_fc;
	char mode;
	bool bool_result;
	bool is_client;

	// Preallocate some variables
	allocated_size = 16384 * sizeof(char);
	max_time = 0;
	remote_host = NULL;
	device_name = NULL;
	is_client = false;

	srand48(getpid() * time(NULL));
	bool_result = common_hashtable_initialize(&mr_mgr, 33);
	if(!bool_result) {
		fprintf(stderr, "Creation of the MR manager failed\n");
		return -1;
	}

	/******************************
	 **** Arguments processing ****
	 ******************************/
	if(strcmp(argv[1], "file") == 0) {
		mode = RDMA_TEST_FILE;
	} else if(strcmp(argv[1], "stress") == 0) {
		mode = RDMA_TEST_STRESS;
	} else {
		fprintf(stderr, "Invalid command, allowed modes are: file and stress\n");
		return -1;
	}

	while((opt = getopt(argc, argv, "c:d:")) != -1) {
		// Detect if a server was given; if any, act as a client
		if(opt == 'c') {
			remote_host_size = strlen(optarg) + 1;
			remote_host = (char *) malloc(remote_host_size * sizeof(char));
			strcpy(remote_host, optarg);

			is_client = true;
		} else if(opt == 'd') {
			device_name_size = strlen(optarg) + 1;
			device_name = (char *) malloc(device_name_size * sizeof(char));
			strcpy(device_name, optarg);
		}
	}

	if(device_name == NULL) {
		fprintf(stderr, "Device name is mandatory: use option `-d`\n");
		return -1;
	}

	// Set up remote address
	if(is_client) {
		memset(&server_hints, 0, sizeof(struct addrinfo));
		server_hints.ai_family = AF_INET6;
		server_hints.ai_socktype = SOCK_DGRAM;

		if(getaddrinfo(remote_host, "7362", &server_hints, &server_infos) != 0) {
			fprintf(stderr, "getaddrinfo error\n");
			return -1;
		}
	}

	/******************************
	 ***** RDMA initialization ****
	 ******************************/
	bool_result = capulet_rdma_ib_initialize_device(&rdma_ctx, device_name);
	if(!bool_result) {
		fprintf(stderr, "Device initialization failed\n");
		return -1;
	}

	in_mr = capulet_rdma_ib_create_mr(&rdma_ctx, allocated_size, IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
	out_mr = capulet_rdma_ib_create_mr(&rdma_ctx, allocated_size, IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);

	if(!in_mr || !out_mr) {
		fprintf(stderr, "Memory registration failed\n");
		return -1;
	}

	bool_result = common_hashtable_insert(&mr_mgr, "in", (void *) in_mr);
	if(!bool_result) {
		fprintf(stderr, "Adding MR to hashtable failed\n");
		return -1;
	}

	bool_result = common_hashtable_insert(&mr_mgr, "out", (void *) out_mr);
	if(!bool_result) {
		fprintf(stderr, "Adding MR to hashtable failed\n");
		return -1;
	}

	bool_result = capulet_rdma_ib_initialize_qp(&rdma_ctx, IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_WRITE);
	if(!bool_result) {
		fprintf(stderr, "Queue Pair initialization failed\n");
		return -1;
	}


	/******************************
	 **** Exchange informations ***
	 ******************************/
	udp_ctx = capulet_net_udp_initialize();
	if(!udp_ctx) {
		fprintf(stderr, "UDP context initialization failed\n");
		return -1;
	}

	bool_result = capulet_rdma_ib_fill_base_udp(&rdma_ctx, udp_ctx->local);
	if(!bool_result) {
		fprintf(stderr, "Query port failed\n");
		return -1;
	}

	capulet_net_udp_dump_qp_packet(udp_ctx->local);

	if(is_client) {
		server_socket = capulet_net_udp_send_qp(udp_ctx, server_infos);
	} else {
		server_socket = capulet_net_udp_serve_qp(udp_ctx);
	}

	if(server_socket == -1) {
		fprintf(stderr, "Connection to peer failed\n");
		return -1;
	}

	capulet_net_udp_dump_qp_packet(udp_ctx->remote);

	bool_result = capulet_rdma_ib_set_peer_from_udp(&rdma_ctx, udp_ctx);
	if(!bool_result) {
		perror("Setting state to RTR/RTS failed");
		return -1;
	}

	if(mode == RDMA_TEST_FILE) {
		/******************************
		 ******** Read or wait ********
		 ******************************/
		if(is_client) {
			read_count = read(STDIN_FILENO, (char *) in_mr->addr, allocated_size);
		} else {
			read_count = read(STDIN_FILENO, (char *) out_mr->addr, allocated_size);
		}

		if(read_count == -1) {
			perror("Reading file failed\n");
			return -1;
		}

		bool_result = capulet_rdma_ib_post_recv(&rdma_ctx, in_mr, allocated_size, 0);
		if(!bool_result) {
			fprintf(stderr, "Posting Recv (in) failed\n");
			return -1;
		}

		bool_result = capulet_rdma_ib_post_recv(&rdma_ctx, out_mr, allocated_size, 0);
		if(!bool_result) {
			fprintf(stderr, "Posting Recv (out) failed\n");
			return -1;
		}

		if(!is_client) {
			serve_mr_td_ctx.mr_mgr = &mr_mgr;
			serve_mr_td_ctx.server_socket = server_socket;

			pthread_create(&serve_mr_td, NULL, *serve_mr_td_fn, (void *) &serve_mr_td_ctx);
		} else {
			bool_result = capulet_net_udp_query_mr(server_socket, server_infos, "in", &in_info_packet);
			if(!bool_result || (strcmp(in_info_packet.name, "in") != 0)) {
				fprintf(stderr, "Querying MR (in) from server failed\n");
				return -1;
			}

			bool_result = capulet_net_udp_query_mr(server_socket, server_infos, "out", &out_info_packet);
			if(!bool_result || (strcmp(out_info_packet.name, "out") != 0)) {
				fprintf(stderr, "Querying MR (out) from server failed\n");
				return -1;
			}

			bool_result = capulet_rdma_ib_post_send(&rdma_ctx, IBV_WR_RDMA_READ, out_mr, allocated_size, 0, &out_info_packet);
			if(!bool_result) {
				fprintf(stderr, "Sending Read failed\n");
				return -1;
			}

			bool_result = capulet_rdma_ib_post_send(&rdma_ctx, IBV_WR_RDMA_WRITE, in_mr, allocated_size, 0, &in_info_packet);
			if(!bool_result) {
				fprintf(stderr, "Sending Write failed\n");
				return -1;
			}
		}

		/******************************
		 ****** Poll completion *******
		 ******************************/
		if(is_client) {
			do {
				poll_result = ibv_poll_cq(rdma_ctx.send_cq, 1, &poll_wc);
				usleep(100);
			} while(poll_result == 0);

			if(poll_result > 0 && poll_wc.status == IBV_WC_SUCCESS) {
				if(is_client) {
					printf("%s\n", (char *) out_mr->addr);
				}
			} else {
				printf("Failed: %s (WR %lu)\n", ibv_wc_status_str(poll_wc.status), poll_wc.wr_id);
			}
		} else {
			// RDMA Read and Write do not give any information on completion
			// on the server side, so we wait for the first char of the string
			// to be non-zero.
			do {
				in_mr_fc = *((char *) in_mr->addr);
				usleep(100);
			} while(in_mr_fc == 0);

			printf("%s\n", (char *) in_mr->addr);
		}
	} else if(mode == RDMA_TEST_STRESS) {
		// Fill the Recv queue
		for(int i = 0; i < 16; i++) {
			capulet_rdma_ib_post_recv(&rdma_ctx, in_mr, allocated_size, 0);
		}

		// A small delay is needed to avoid sending anything before Recv requests
		// are posted on the other side.
		usleep(1000);

		while(1) {
			clock_gettime(CLOCK_MONOTONIC, &timestamps[0]);

			// Wait for Send
			capulet_rdma_ib_post_send(&rdma_ctx, IBV_WR_SEND, out_mr, allocated_size, 0, NULL);
			do {
				poll_result = ibv_poll_cq(rdma_ctx.send_cq, 1, &poll_wc);
			} while(poll_result == 0);

			if(poll_result < 0 || poll_wc.status != IBV_WC_SUCCESS) {
				printf("WR %lu failed: %s (%d)\n", poll_wc.wr_id, ibv_wc_status_str(poll_wc.status), poll_wc.vendor_err);
				return -1;
			}

			// Wait for Recv
			do {
				poll_result = ibv_poll_cq(rdma_ctx.recv_cq, 1, &poll_wc);
			} while(poll_result == 0);

			if(poll_result < 0 || poll_wc.status != IBV_WC_SUCCESS) {
				printf("WR %lu failed: %s (%d)\n", poll_wc.wr_id, ibv_wc_status_str(poll_wc.status), poll_wc.vendor_err);
				return -1;
			}

			// The Recv queue should always be full for optimal performances: once
			// an element is consumed, push back to it.
			capulet_rdma_ib_post_recv(&rdma_ctx, in_mr, allocated_size, 0);
			clock_gettime(CLOCK_MONOTONIC, &timestamps[1]);

			elapsed_time = (timestamps[1].tv_sec - timestamps[0].tv_sec) * 1000000000 + timestamps[1].tv_nsec - timestamps[0].tv_nsec;
			if(elapsed_time > max_time) max_time = elapsed_time;
			printf("\r%8lu / %8lu", elapsed_time, max_time);
			fflush(stdout);
		}
	}

	/******************************
	 ******* Global cleanup *******
	 ******************************/
	capulet_net_udp_free(udp_ctx);
	capulet_rdma_ib_free_mr(in_mr);
	capulet_rdma_ib_free_mr(out_mr);
	common_hashtable_free(&mr_mgr);
	capulet_rdma_ib_free(&rdma_ctx);
	freeaddrinfo(server_infos);

	return 0;
}

