#include "libcapulet/net_udp.h"

struct CapuletNetUdpContext *capulet_net_udp_initialize() {
	struct CapuletNetUdpContext *ctx;
	void *raw_ptr;

	const size_t packet_size = sizeof(struct CapuletNetUdpQpInfoPacket);
	const size_t ctx_size = sizeof(struct CapuletNetUdpContext);
	const size_t total_size = 2 * packet_size + ctx_size;

	raw_ptr = malloc(total_size);
	if(!raw_ptr) {
		return NULL;
	}

	memset(raw_ptr, 0, total_size);

	ctx = (struct CapuletNetUdpContext *) raw_ptr;
	ctx->local = (struct CapuletNetUdpQpInfoPacket *) (raw_ptr + ctx_size);
	ctx->remote = (struct CapuletNetUdpQpInfoPacket *) (raw_ptr + ctx_size + packet_size);

	return ctx;	
}

int capulet_net_udp_serve_qp(struct CapuletNetUdpContext *ctx) {
	struct addrinfo server_hints;
	struct addrinfo *server_infos;
	struct sockaddr_storage client_addr;

	int result;
	int server_socket;

	size_t buffer_size = sizeof(struct CapuletNetUdpQpInfoPacket);
	uint8_t receive_buffer[buffer_size];

	socklen_t client_addr_size = sizeof(struct sockaddr_storage);

	// Set up bind address
	memset(&server_hints, 0, sizeof(struct addrinfo));
	server_hints.ai_family = AF_INET6;
	server_hints.ai_socktype = SOCK_DGRAM;
	server_hints.ai_flags = AI_PASSIVE;

	result = getaddrinfo(NULL, "7362", &server_hints, &server_infos);
	if(result != 0) {
		fprintf(stderr, "getaddrinfo error: %s\n", gai_strerror(result));
		return -1;
	}

	// Create a socket an bind
	// XXX: assumes first address in linked list is 0.0.0.0
	server_socket = socket(server_infos->ai_family, server_infos->ai_socktype, server_infos->ai_protocol);
	if(server_socket == -1) {
		perror("socket error");
		return -1;
	}

	result = bind(server_socket, server_infos->ai_addr, server_infos->ai_addrlen);
	if(result == -1) {
		perror("bind error");
		return -1;
	}

	// Address info can be freed as soon as the socket in bound
	freeaddrinfo(server_infos);

	recvfrom(server_socket, receive_buffer, buffer_size, 0, (struct sockaddr *) &client_addr, &client_addr_size);
	memcpy((void *) ctx->remote, (void *) receive_buffer, buffer_size);
	sendto(server_socket, (void *) ctx->local, buffer_size, MSG_CONFIRM, (struct sockaddr *) &client_addr, client_addr_size);

	return server_socket;
}

bool capulet_net_udp_serve_mr(struct CommonHashtableTable *hashtable, int server_socket) {
	struct sockaddr_storage client_addr;
	socklen_t client_addr_size = sizeof(struct sockaddr_storage);

	struct CapuletNetUdpMrQueryPacket mr_query_pkt;
	struct CapuletNetUdpMrInfoPacket mr_info_pkt;
	struct CommonHashtableElement *el;
	struct ibv_mr *mr;

	ssize_t recv_size;
	size_t buffer_size;
	uint8_t receive_buffer[256];

	while(1) {
		// No need to go too fast, and CPU time is better used anywhere else
		usleep(10);

		buffer_size = sizeof(struct CapuletNetUdpMrQueryPacket);
		recv_size = recvfrom(server_socket, receive_buffer, buffer_size, 0, (struct sockaddr *) &client_addr, &client_addr_size);
		memcpy((void *) &mr_query_pkt, (void *) receive_buffer, buffer_size);

		buffer_size = sizeof(struct CapuletNetUdpMrInfoPacket);
		memset(&mr_info_pkt, 0, buffer_size);

		if(recv_size == -1) return false;
		el = common_hashtable_find(hashtable, mr_query_pkt.name);

		if(el) {
			mr = (struct ibv_mr *) el->content;

			memcpy((void *) mr_info_pkt.name, (void *) el->name, 16);
			mr_info_pkt.addr = mr->addr;
			mr_info_pkt.size = mr->length;
			mr_info_pkt.key = mr->rkey;

			sendto(server_socket, (void *) &mr_info_pkt, buffer_size, MSG_CONFIRM, (struct sockaddr *) &client_addr, client_addr_size);
		}

	}

	close(server_socket);

	return true;
}

int capulet_net_udp_send_qp(struct CapuletNetUdpContext *ctx, struct addrinfo *server_infos) {
	int server_socket;
	int result;

	const size_t buffer_size = sizeof(struct CapuletNetUdpQpInfoPacket);
	uint8_t receive_buffer[buffer_size];

	// Create a socket
	// XXX: assumes first address in linked list is 0.0.0.0
	server_socket = socket(server_infos->ai_family, server_infos->ai_socktype, server_infos->ai_protocol);
	if(server_socket == -1) {
		perror("socket error");
		return -1;
	}

	result = sendto(server_socket, (void *) ctx->local, buffer_size, 0, server_infos->ai_addr, server_infos->ai_addrlen);
	if(result == -1) {
		perror("sendto error");
		return -1;
	}

	// Wait for informations from the server
	result = recvfrom(server_socket, receive_buffer, buffer_size, 0, server_infos->ai_addr, &server_infos->ai_addrlen);
	if(result == -1) {
		perror("recvfrom error");
		return -1;
	}

	memcpy((void *) ctx->remote, (void *) receive_buffer, buffer_size);

	return server_socket;
}

bool capulet_net_udp_query_mr(int server_socket, struct addrinfo *server_infos, char *name, struct CapuletNetUdpMrInfoPacket *mr_info_packet) {
	struct CapuletNetUdpMrQueryPacket mr_query_packet;

	size_t buffer_size = sizeof(struct CapuletNetUdpMrInfoPacket);
	int result;
	uint8_t receive_buffer[buffer_size];

	// Ensure string will always be NUL-terminated
	strncpy((void *) mr_query_packet.name, (void *) name, 15);
	mr_query_packet.name[15] = '\0';

	result = sendto(server_socket, (void *) &mr_query_packet, sizeof(struct CapuletNetUdpMrQueryPacket), 0, server_infos->ai_addr, server_infos->ai_addrlen);
	if(result == -1) {
		perror("sendto error");
		return false;
	}

	// Wait for informations from the server
	result = recvfrom(server_socket, receive_buffer, buffer_size, 0, server_infos->ai_addr, &server_infos->ai_addrlen);
	if(result == -1) {
		perror("recvfrom error");
		return false;
	}

	memcpy((void *) mr_info_packet, (void *) receive_buffer, buffer_size);

	return true;
}

void capulet_net_udp_dump_qp_packet(struct CapuletNetUdpQpInfoPacket *packet) {
	printf("LID: 0x%x | QPN: 0x%x | PSN: 0x%x | GID: ", packet->lid, packet->qpn, packet->psn);

	for(char i = 0; i < 8; i++) {
		printf("%x%x ", packet->gid[2 * i], packet->gid[2 * i + 1]);
	}

	printf("\n");
}

void capulet_net_udp_free(struct CapuletNetUdpContext *ctx) {
	free((void *) ctx);
}

