#include "libcapulet/rdma_ib.h"

bool capulet_rdma_ib_initialize_device(struct CapuletRdmaIbContext *ctx, const char *device_name) {
	struct ibv_device **dev_list;
	struct ibv_device *ibv_dev = NULL;

	int dev_list_length, dev_id;

	// Devices cannot be selected by name, so all devices are listed
	// and the name of each one is compared to the requested name.
	dev_list = ibv_get_device_list(&dev_list_length);
	if(!dev_list || dev_list_length < 1) {
		return false;
	}

	for(dev_id = 0; dev_id < dev_list_length; dev_id++) {
		const char *current_name = ibv_get_device_name(dev_list[dev_id]);
		if(strcmp(device_name, current_name) == 0) {
			ibv_dev = dev_list[dev_id];
		}
	}
	
	if(!ibv_dev) {
		return false;
	}

	// The devices list can be freed as soon as context was obtained.
	ctx->ib_ctx = ibv_open_device(ibv_dev);
	ibv_free_device_list(dev_list);

	// XXX: for now, a global Protection Domain is allocated
	ctx->pd = ibv_alloc_pd(ctx->ib_ctx);
	if(!ctx->pd) {
		return false;
	}

	return true;
}

bool capulet_rdma_ib_initialize_qp(struct CapuletRdmaIbContext *ctx, int access) {
	struct ibv_qp_init_attr ibv_dev_qp_request;
	struct ibv_qp_attr ibv_dev_qp_params;
	struct ibv_qp *ibv_dev_qp;

	int result;

	ctx->send_cq = ibv_create_cq(ctx->ib_ctx, 32, NULL, NULL, 0);
	ctx->recv_cq = ibv_create_cq(ctx->ib_ctx, 32, NULL, NULL, 0);
	if(!ctx->send_cq || !ctx->recv_cq) {
		return false;
	}

	memset(&ibv_dev_qp_request, 0, sizeof(struct ibv_qp_init_attr));
	memset(&ibv_dev_qp_params, 0, sizeof(struct ibv_qp_attr));

	// Two steps are required: creating the queue pair and
	// initializing it.
	// Creation of QP with initial parameters
	ibv_dev_qp_request.qp_type = IBV_QPT_RC;
	ibv_dev_qp_request.send_cq = ctx->send_cq;
	ibv_dev_qp_request.recv_cq = ctx->recv_cq;

	ibv_dev_qp_request.cap.max_send_wr = 32;
	ibv_dev_qp_request.cap.max_recv_wr = 32;
	ibv_dev_qp_request.cap.max_send_sge = 32;
	ibv_dev_qp_request.cap.max_recv_sge = 32;

	ibv_dev_qp = ibv_create_qp(ctx->pd, &ibv_dev_qp_request);

	if(!ibv_dev_qp) {
		return false;
	}

	// Definition of QP state to "Initialized" and access control
	ibv_dev_qp_params.qp_state = IBV_QPS_INIT;
	ibv_dev_qp_params.pkey_index = 0;
	ibv_dev_qp_params.port_num = 1;
	ibv_dev_qp_params.qp_access_flags = access;

	result = ibv_modify_qp(ibv_dev_qp, &ibv_dev_qp_params, IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_PORT | IBV_QP_ACCESS_FLAGS);
	if(result) {
		return false;
	}

	ctx->qp = ibv_dev_qp;

	return true;
}

struct ibv_mr *capulet_rdma_ib_create_mr(struct CapuletRdmaIbContext *ib_ctx, size_t size, int flags) {
	struct ibv_mr *memory_buffer_mr;
	void *memory_buffer;
	int page_size = sysconf(_SC_PAGESIZE);

	// Allocate a memory buffer
	memory_buffer = aligned_alloc(page_size, size);
	if(!memory_buffer) {
		return NULL;
	}

	memset(memory_buffer, 0, size);

	// Register a RDMA MR
	memory_buffer_mr = ibv_reg_mr(ib_ctx->pd, memory_buffer, size, flags);
	if(!memory_buffer_mr) {
		return NULL;
	}

	return memory_buffer_mr;
}

bool capulet_rdma_ib_fill_base_udp(struct CapuletRdmaIbContext *ctx, struct CapuletNetUdpQpInfoPacket *own) {
	struct ibv_port_attr ibv_dev_attr;
	union ibv_gid ibv_dev_gid;

	int result;

	result = ibv_query_port(ctx->ib_ctx, 1, &ibv_dev_attr);
	if(result) {
		return false;
	}

	result = ibv_query_gid(ctx->ib_ctx, 1, 0, &ibv_dev_gid);
	if(result) {
		return false;
	}

	memcpy((void *) own->gid, (void *) ibv_dev_gid.raw, 16);

	own->lid = ibv_dev_attr.lid;
	own->qpn = ctx->qp->qp_num;
	own->psn = lrand48() & 0xffffff;

	return true;	
}

bool capulet_rdma_ib_set_peer_from_udp(struct CapuletRdmaIbContext *ctx, struct CapuletNetUdpContext *udp_ctx) {
	struct ibv_qp_attr ibv_dev_qp_params;
	union ibv_gid ibv_dev_gid;

	int result;

	// Insert GID into IBv structure
	memcpy((void *) ibv_dev_gid.raw, (void *) udp_ctx->remote->gid, 16);

	// Change state from Init to RTR (Ready To Receive)
	memset(&ibv_dev_qp_params, 0, sizeof(struct ibv_qp_attr));
	ibv_dev_qp_params.qp_state = IBV_QPS_RTR;
	ibv_dev_qp_params.path_mtu = IBV_MTU_1024;
	ibv_dev_qp_params.dest_qp_num = udp_ctx->remote->qpn;
	ibv_dev_qp_params.rq_psn = udp_ctx->remote->psn;
	ibv_dev_qp_params.max_dest_rd_atomic = 1;
	// 0xd = 960 μs delay
	ibv_dev_qp_params.min_rnr_timer = 0xd;

	ibv_dev_qp_params.ah_attr.is_global = 1;
	ibv_dev_qp_params.ah_attr.dlid = udp_ctx->remote->lid;
	ibv_dev_qp_params.ah_attr.sl = 0;
	ibv_dev_qp_params.ah_attr.src_path_bits = 0;
	ibv_dev_qp_params.ah_attr.port_num = 1;

	ibv_dev_qp_params.ah_attr.grh.hop_limit = 1;
	ibv_dev_qp_params.ah_attr.grh.dgid = ibv_dev_gid;
	ibv_dev_qp_params.ah_attr.grh.sgid_index = 0;

	result = ibv_modify_qp(ctx->qp, &ibv_dev_qp_params, IBV_QP_STATE | IBV_QP_AV | IBV_QP_PATH_MTU | IBV_QP_DEST_QPN | IBV_QP_RQ_PSN | IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER);
	if(result) {
		return false;
	}

	// Change state from RTR to RTS (Ready To Send)
	memset(&ibv_dev_qp_params, 0, sizeof(struct ibv_qp_attr));
	ibv_dev_qp_params.qp_state = IBV_QPS_RTS;
	// Debugging parameters: infinite timeout
	// This will make the polling wait forever when a race condition occurs
	ibv_dev_qp_params.timeout = 0;
	ibv_dev_qp_params.retry_cnt = 2;
	ibv_dev_qp_params.rnr_retry = 2;
	ibv_dev_qp_params.sq_psn = udp_ctx->local->psn;
	ibv_dev_qp_params.max_rd_atomic = 1;

	result = ibv_modify_qp(ctx->qp, &ibv_dev_qp_params, IBV_QP_STATE | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT | IBV_QP_RNR_RETRY | IBV_QP_SQ_PSN | IBV_QP_MAX_QP_RD_ATOMIC);
	if(result) {
		return false;
	}

	return true;
}

bool capulet_rdma_ib_post_recv(struct CapuletRdmaIbContext *ctx, struct ibv_mr *ibv_dev_mr, int mr_size, uint32_t id_prefix) {
	struct ibv_sge ibv_dev_sge;
	struct ibv_recv_wr ibv_dev_rdma_wr;
	struct ibv_recv_wr *ibv_dev_bad_wr = NULL;

	int result;

	memset(&ibv_dev_sge, 0, sizeof(struct ibv_sge));
	memset(&ibv_dev_rdma_wr, 0, sizeof(struct ibv_recv_wr));

	ibv_dev_sge.addr = (uintptr_t) ibv_dev_mr->addr;
	ibv_dev_sge.length = mr_size;
	ibv_dev_sge.lkey = ibv_dev_mr->lkey;

	ibv_dev_rdma_wr.wr_id = (((uint64_t) id_prefix) << 32) | ((uint64_t) lrand48());
	ibv_dev_rdma_wr.sg_list = &ibv_dev_sge;
	ibv_dev_rdma_wr.num_sge = 1;

	result = ibv_post_recv(ctx->qp, &ibv_dev_rdma_wr, &ibv_dev_bad_wr);
	if(result) {
		perror("Could not post Recv request");
		return false;
	}

	return ibv_dev_bad_wr == NULL;
}

bool capulet_rdma_ib_post_send(struct CapuletRdmaIbContext *ctx, enum ibv_wr_opcode opcode, struct ibv_mr *local_mr, int local_mr_size, uint32_t id_prefix, struct CapuletNetUdpMrInfoPacket *remote_mr) {
	struct ibv_send_wr ibv_dev_rdma_wr;
	struct ibv_sge ibv_dev_sge;
	struct ibv_send_wr *ibv_dev_bad_wr = NULL;

	int result;

	memset(&ibv_dev_sge, 0, sizeof(struct ibv_sge));
	memset(&ibv_dev_rdma_wr, 0, sizeof(struct ibv_send_wr));

	ibv_dev_sge.addr = (uintptr_t) local_mr->addr;
	ibv_dev_sge.length = local_mr_size;
	ibv_dev_sge.lkey = local_mr->lkey;

	ibv_dev_rdma_wr.wr_id = (((uint64_t) id_prefix) << 32) | ((uint64_t) lrand48());
	ibv_dev_rdma_wr.sg_list = &ibv_dev_sge;
	ibv_dev_rdma_wr.num_sge = 1;
	ibv_dev_rdma_wr.opcode = opcode;
	ibv_dev_rdma_wr.send_flags = IBV_SEND_SIGNALED;

	if(opcode == IBV_WR_RDMA_READ || opcode == IBV_WR_RDMA_WRITE) {
		ibv_dev_rdma_wr.wr.rdma.remote_addr = (uintptr_t) remote_mr->addr;
		ibv_dev_rdma_wr.wr.rdma.rkey = remote_mr->key;
	}

	result = ibv_post_send(ctx->qp, &ibv_dev_rdma_wr, &ibv_dev_bad_wr);

	if(result) {
		perror("Could not post Send request");
		return false;
	}

	return ibv_dev_bad_wr == NULL;
}

void capulet_rdma_ib_free_mr(struct ibv_mr *mr) {
	free(mr->addr);
	ibv_dereg_mr(mr);
}

void capulet_rdma_ib_free(struct CapuletRdmaIbContext *ctx) {
	if(ctx->send_cq && ctx->recv_cq) {
		ibv_destroy_cq(ctx->send_cq);
		ibv_destroy_cq(ctx->recv_cq);
	}

	if(ctx->qp) ibv_destroy_qp(ctx->qp);

	ibv_dealloc_pd(ctx->pd);
	ibv_close_device(ctx->ib_ctx);
}

