#include "libpotoml/toml.h"

char potoml_toml_parse(struct CommonHashtableTable *hashtable, char *toml) {
	size_t toml_length = strlen(toml);

	size_t section_length = 0;
	size_t property_length = 0;
	size_t value_length = 0;
	size_t ci;

	char c;
	char s = PotomlLineBeginning;

	char *section = malloc(POTOML_MAX_PROPERTY_LENGTH * sizeof(char));
	char *property = malloc(POTOML_MAX_PROPERTY_LENGTH * sizeof(char));
	char *value = malloc(POTOML_MAX_PROPERTY_LENGTH * sizeof(char));

	void *casted_value;

	for(ci = 0; ci < toml_length; ci++) {
		c = toml[ci];

		switch(s) {
			case PotomlLineBeginning:
				if(c == '#') s = PotomlCommentContent;
				else if(c == '[') {
					memset(section, 0, POTOML_MAX_PROPERTY_LENGTH);
					section_length = 0;

					s = PotomlSectionName;
				} else if((c >= 65 && c <= 90) || (c >= 97 && c <= 122)) {
					memset(property, 0, POTOML_MAX_PROPERTY_LENGTH);
					property_length = 0;
					value = malloc(POTOML_MAX_PROPERTY_LENGTH * sizeof(char));
					memset(value, 0, POTOML_MAX_PROPERTY_LENGTH);
					value_length = 0;

					// Append section name before property for storing in hashtable
					strncpy(property, section, section_length);
					strcat(property, ".");

					ci--;
					s = PotomlPropertyName;
				}
				else if(c != '\n' && c != ' ' && c != '\t') return InvalidCharacter;
				break;

			case PotomlLineEnding:
				if(c == '\n') s = PotomlLineBeginning;
				else if(c != ' ' && c != '\t') return InvalidCharacter;
				break;

			case PotomlCommentContent:
				if(c == '\n') s = PotomlLineBeginning;
			 	break;

			case PotomlSectionName:
				if(c == ']') {
					s = PotomlLineEnding;
					break;
				}
				else if(c == '\n') return MissingClosingBracket;

				section[section_length] = c;
				section_length++;

				break;

			case PotomlPropertyName:
				if(c == ' ' || c == '\t') s = PotomlAfterPropertyName;
				else if(c == '=') s = PotomlBeforePropertyValue;
				else {
					property[section_length + property_length + 1] = c;
					property_length++;
				}

				break;

			case PotomlAfterPropertyName:
				if(c == '=') s = PotomlBeforePropertyValue;
				else if(c != ' ' && c != '\t') return InvalidCharacter;
				break;

			case PotomlBeforePropertyValue:
				if(c == '"') s = PotomlPropertyValueString;
				else if(c >= 48 && c <= 57) {
					ci--;
					s = PotomlPropertyValueLong;
				} else if(c != ' ' && c != '\t') return InvalidCharacter;
				break;

			case PotomlPropertyValueString:
				if(c == '"') {
					if(!common_hashtable_insert(hashtable, property, value)) {
						return HashtableInsertError;
					}

					s = PotomlLineEnding;
					break;
				}

				value[value_length] = c;
				value_length++;
				break;

			case PotomlPropertyValueLong:
			case PotomlPropertyValueDouble:
				if(c == '\n' || c == ' ' || c == '\t') {
					if(s == PotomlPropertyValueLong) {
						casted_value = malloc(sizeof(long));
						*((long *) casted_value) = strtol(value, NULL, 10);
					} else /* if(s == PotomlPropertyValueDouble) */ {
						casted_value = malloc(sizeof(double));
						*((double *) casted_value) = strtod(value, NULL);
					}

					if(!common_hashtable_insert(hashtable, property, casted_value)) {
						return HashtableInsertError;
					}

					if(c == '\n') ci--;
					s = PotomlLineEnding;
					break;
				} else if(c == '.') {
					if(s == PotomlPropertyValueLong) s = PotomlPropertyValueDouble;
					else return InvalidCharacter;
				} else if(c < 48 || c > 57) return InvalidCharacter;

				value[value_length] = c;
				value_length++;
				break;
		}
	}

	free(section);
	free(property);

	return 0;
}

char potoml_toml_parse_file(struct CommonHashtableTable *hashtable, char *toml_file_path) {
	FILE *toml_file_handle;
	char *toml_file_content;
	size_t toml_file_length;
	size_t result;

	toml_file_handle = fopen(toml_file_path, "r");
	if(!toml_file_handle) {
		return FileDoesNotExist;
	}

	// Allocate the right size
	fseek(toml_file_handle, 0, SEEK_END);
	toml_file_length = ftell(toml_file_handle);
	rewind(toml_file_handle);

	toml_file_content = malloc((toml_file_length + 1) * sizeof(char));
	result = fread(toml_file_content, toml_file_length, 1, toml_file_handle);

	if(result != 1) {
		return CouldNotReadFromFile;
	}

	char out = potoml_toml_parse(hashtable, toml_file_content);

	free(toml_file_content);
	fclose(toml_file_handle);

	return out;
}

struct PotomlTomlSubset *potoml_toml_create_subset(struct CommonHashtableTable *hashtable, const char *section_name) {
	struct PotomlTomlSubset *subset;

	size_t section_name_length = strlen(section_name);
	char *owned_section_name = malloc((section_name_length + 2) * sizeof(char));

	if(!owned_section_name) return NULL;
	strcpy(owned_section_name, section_name);

	// Append dot to section name if non-empty
	if(section_name_length != 0) strcat(owned_section_name, ".");

	subset = (struct PotomlTomlSubset *) malloc(sizeof(struct PotomlTomlSubset));

	if(!subset) return NULL;

	subset->hashtable = hashtable;
	subset->section_name = owned_section_name;
	subset->section_name_length = section_name_length + 1;

	return subset;
}

char *potoml_toml_get_string(struct PotomlTomlSubset *subset, const char *property) {
	struct CommonHashtableElement *el;

	size_t property_length;
	char *full_property;

	property_length = strlen(property);

	full_property = (char *) malloc((property_length + subset->section_name_length + 1) * sizeof(char));
	strcpy(full_property, subset->section_name);
	strcat(full_property, property);

	el = common_hashtable_find(subset->hashtable, full_property);
	if(!el) {
		return NULL;
	}

	free(full_property);
	return (char *) el->content;
}

double potoml_toml_get_double(struct PotomlTomlSubset *subset, const char *property) {
	struct CommonHashtableElement *el;

	size_t property_length;
	double value;
	char *full_property;

	value = 0.0;
	property_length = strlen(property);

	full_property = (char *) malloc(property_length + subset->section_name_length + 1);
	strcpy(full_property, subset->section_name);
	strcat(full_property, property);

	el = common_hashtable_find(subset->hashtable, full_property);
	if(el) {
		// Clone instead of giving reference for double
		value = *((double *) el->content);
	}

	free(full_property);
	return value;
}

long potoml_toml_get_long(struct PotomlTomlSubset *subset, const char *property) {
	struct CommonHashtableElement *el;

	size_t property_length;
	long value;
	char *full_property;

	value = 0;
	property_length = strlen(property);

	full_property = (char *) malloc(property_length + subset->section_name_length + 1);
	strcpy(full_property, subset->section_name);
	strcat(full_property, property);

	el = common_hashtable_find(subset->hashtable, full_property);
	if(el) {
		// Clone instead of giving reference for long
		value = *((long *) el->content);
	}

	free(full_property);
	return value;
}

void potoml_toml_subset_free(struct PotomlTomlSubset *subset) {
	free((void *) subset->section_name);
	free((void *) subset);
}

void potoml_toml_free(struct CommonHashtableTable *hashtable) {
	unsigned char i;
	struct CommonHashtableElement *el;

	// Free allocated elements 
	for(i = 0; i < hashtable->size; i++) {
		el = hashtable->base + i;
		if(el->content != NULL) free((void *) el->content);
	}

	common_hashtable_free(hashtable);
}

