#include "libtrx/trx_rdma.h"

int trx_rdma_start(TRXState *s, const TRXDriverParams2 *p) {
	struct SDRContext *sdr_context;

	struct addrinfo server_hints;
	struct addrinfo *server_infos;

	int server_socket;
	int mr_id;
	bool result;

	sdr_context = s->opaque;

	// RDMA is used in Send/Recv mode so intrinsically it does not have a notion of client or
	// server. However, parameters are exchanged over UDP, and having a server is convenient.
	// If acting as a client, the remote address should therefore be set up.
	if(sdr_context->server_addr) {
		memset(&server_hints, 0, sizeof(struct addrinfo));
		server_hints.ai_family = AF_INET6;
		server_hints.ai_socktype = SOCK_DGRAM;
		if(getaddrinfo(sdr_context->server_addr, "7362", &server_hints, &server_infos) != 0) return -1;
	}

	// Get IBV context for device
	result = capulet_rdma_ib_initialize_device(&sdr_context->ib_ctx, sdr_context->device_name);
	if(!result) return -1;

	for(mr_id = 0; mr_id < TRX_RDMA_MR_COUNT; mr_id++) {
		sdr_context->recv_mrs[mr_id] = capulet_rdma_ib_create_mr(&sdr_context->ib_ctx, sizeof(struct SDRMemoryRegion), IBV_ACCESS_LOCAL_WRITE);
		sdr_context->send_mrs[mr_id] = capulet_rdma_ib_create_mr(&sdr_context->ib_ctx, sizeof(struct SDRMemoryRegion), IBV_ACCESS_LOCAL_WRITE);
		if(!sdr_context->recv_mrs[mr_id] || !sdr_context->send_mrs[mr_id]) return -1;
	}

	result = capulet_rdma_ib_initialize_qp(&sdr_context->ib_ctx, IBV_ACCESS_LOCAL_WRITE);
	if(!result) return -1;

	// Create UDP context and exchange informations with peer
	// Sends three packets: one QP informations and two MR queries
	sdr_context->udp_ctx = capulet_net_udp_initialize();
	if(!sdr_context->udp_ctx) return -1;

	result = capulet_rdma_ib_fill_base_udp(&sdr_context->ib_ctx, sdr_context->udp_ctx->local);
	if(!result) return -1;

	// Server must wait for data while client will send them
	if(sdr_context->server_addr) {
		server_socket = capulet_net_udp_send_qp(sdr_context->udp_ctx, server_infos);
	} else {
		server_socket = capulet_net_udp_serve_qp(sdr_context->udp_ctx);
	}
	if(!server_socket) return -1;

	// Set state to RTS, does not require any MR informations
	result = capulet_rdma_ib_set_peer_from_udp(&sdr_context->ib_ctx, sdr_context->udp_ctx);
	if(!result) return -1;

	// Fill the Recv queue: initially add one request for each MR
	for(mr_id = 0; mr_id < TRX_RDMA_MR_COUNT; mr_id++) {
		capulet_rdma_ib_post_recv(&sdr_context->ib_ctx, sdr_context->recv_mrs[mr_id], sizeof(struct SDRMemoryRegion), mr_id);
	}
	usleep(10000);

	sdr_context->recv_buffer = common_circular_buffer_create(TRX_RDMA_BUFFER_SIZE * sizeof(TRXComplex), false);

	return 0;
}

void trx_rdma_write(TRXState *s, trx_timestamp_t timestamp, const void **samples, int count, int tx_port_index, TRXWriteMetadata *md) {
	struct SDRContext *sdr_context;
	struct SDRMemoryRegion *out_data;

	TRXComplex **channels;

	struct ibv_wc poll_wc[TRX_RDMA_MR_COUNT];

	int total_written;
	int written;
	int result;
	int mr_id;
	int i;

	sdr_context = s->opaque;
	channels = (TRXComplex **) samples;
	total_written = 0;

	// First write given bytes to a MR
	while(total_written < count) {
		// Find a new MR if none was already filled in
		if(sdr_context->consumed_mr_id == -1) {
			// First set bit (lowest first) corresponds to the first MR available
			// Make sure the bit can be found (ie. chack range) to avoid UB
			if(sdr_context->send_mrs_status > 0 && sdr_context->send_mrs_status < ((uint64_t) 1 << TRX_RDMA_MR_COUNT)) {
				sdr_context->consumed_mr_id = __builtin_ctzl(sdr_context->send_mrs_status);
				sdr_context->consumed_mr_length = 0;
				sdr_context->send_mrs_status &= ~((uint64_t) 1 << sdr_context->consumed_mr_id);
			} else break;
		}

		// Write as many samples as possible to the MR
		written = MIN(count - total_written, TRX_RDMA_MR_SIZE - sdr_context->consumed_mr_length);
		out_data = (struct SDRMemoryRegion *) sdr_context->send_mrs[sdr_context->consumed_mr_id]->addr;

		memcpy((void *) (out_data->iq[0] + sdr_context->consumed_mr_length), (void *) (channels[0] + total_written), written * sizeof(TRXComplex));
		memcpy((void *) (out_data->iq[1] + sdr_context->consumed_mr_length), (void *) (channels[1] + total_written), written * sizeof(TRXComplex));

		total_written += written;
		sdr_context->consumed_mr_length += written;
		// When an MR is full, send it, next iteration will find another one
		if(sdr_context->consumed_mr_length == TRX_RDMA_MR_SIZE) {
			out_data->meta.sample_count = TRX_RDMA_MR_SIZE;
			out_data->meta.frame_number = sdr_context->frame_count++;

			capulet_rdma_ib_post_send(&sdr_context->ib_ctx, IBV_WR_SEND, sdr_context->send_mrs[sdr_context->consumed_mr_id], sizeof(struct SDRMemoryRegion), sdr_context->consumed_mr_id, NULL);
			sdr_context->consumed_mr_id = -1;
		}
	}

	// Regardless of data written, poll for completion of Send requests
	result = ibv_poll_cq(sdr_context->ib_ctx.send_cq, TRX_RDMA_MR_COUNT, poll_wc);
	if(result <= 0) return;

	for(i = 0; i < result; i++) {
		if(poll_wc[i].status != IBV_WC_SUCCESS) {
			printf("(Send) WR %lx failed: %s (%d)\n", poll_wc[i].wr_id, ibv_wc_status_str(poll_wc[i].status), poll_wc[i].vendor_err);
		}

		// MR id is encoded in the WR id prefix (32 first bits over 64 total)
		mr_id = (uint32_t) (poll_wc[i].wr_id >> 32);
		sdr_context->send_mrs_status |= (uint64_t) 1 << mr_id;
	}
}

int trx_rdma_read(TRXState *s, trx_timestamp_t *ptimestamp, void **psamples, int count, int rx_port_index, TRXReadMetadata *md) {
	struct SDRContext *sdr_context;
	struct SDRMemoryRegion *in_data;

	TRXComplex **channels;

	struct ibv_wc poll_wc;

	int result;
	uint32_t mr_id;
	uint32_t byte_count;
	uint16_t sample_count;

	sdr_context = s->opaque;
	channels = (TRXComplex **) psamples;

	if(count <= 0) return 0;

	// Poll for Recv completion
	result = ibv_poll_cq(sdr_context->ib_ctx.recv_cq, 1, &poll_wc);

	if(result < 0) {
		printf("trx_rdma: failed to poll Recv CQ\n");
		return 0;
	}

	if(result > 0) {
		if(poll_wc.status != IBV_WC_SUCCESS) {
			printf("trx_rdma: Recv WR %lx failed: %s (%d)\n", poll_wc.wr_id, ibv_wc_status_str(poll_wc.status), poll_wc.vendor_err);
			return 0;
		}

		mr_id = (uint32_t) (poll_wc.wr_id >> 32);
		in_data = (struct SDRMemoryRegion *) sdr_context->recv_mrs[mr_id]->addr;

		// The Recv queue should always be full for optimal performances: once
		// an element is consumed, push back to it.
		capulet_rdma_ib_post_recv(&sdr_context->ib_ctx, sdr_context->recv_mrs[mr_id], sizeof(struct SDRMemoryRegion), mr_id);

		// Push data onto a circular buffer
		sample_count = in_data->meta.sample_count;
		common_circular_buffer_write(sdr_context->recv_buffer, (void *) in_data->iq[0], sample_count * sizeof(TRXComplex));
	}

	// Copy data from circular buffer to local buffer
	// XXX: handle multiple channels
	byte_count = common_circular_buffer_read(sdr_context->recv_buffer, (void *) channels[0], count * sizeof(TRXComplex));
	if(byte_count != 0) memcpy((void *) channels[1], (void *) channels[0], byte_count);
	sample_count = (uint16_t) (byte_count / sizeof(TRXComplex));

	// Increment timestamp by IQ samples count
	sdr_context->last_timestamp += sample_count;
	*ptimestamp = sdr_context->last_timestamp;

	return sample_count;
}

static void trx_rdma_end(TRXState *s) {
	struct SDRContext *sdr_context;

	int mr_id;

	sdr_context = s->opaque;

	for(mr_id = 0; mr_id < TRX_RDMA_MR_COUNT; mr_id++) {
		if(sdr_context->recv_mrs[mr_id]) capulet_rdma_ib_free_mr(sdr_context->recv_mrs[mr_id]);
		if(sdr_context->send_mrs[mr_id]) capulet_rdma_ib_free_mr(sdr_context->send_mrs[mr_id]);
	}

	capulet_net_udp_free(sdr_context->udp_ctx);
	capulet_rdma_ib_free(&sdr_context->ib_ctx);

	free((void *) sdr_context->device_name);
	if(sdr_context->server_addr) free((void *) sdr_context->server_addr);
}

int trx_driver_init(TRXState *s) {
	struct SDRContext *sdr_context;

	srand48(getpid() * time(NULL));
	sdr_context = calloc(1, sizeof(struct SDRContext));

	if(s->trx_api_version != TRX_API_VERSION) {
		fprintf(stderr, "ABI compatibility mismatch between LTEENB and TRX driver (LTEENB ABI version=%d, TRX driver ABI version=%d)\n", s->trx_api_version, TRX_API_VERSION);
		return -1;
	}

	// String is owned by the driver, and freed in `trx_end_func`
	sdr_context->server_addr = trx_get_param_string(s, "server_addr");
	sdr_context->device_name = trx_get_param_string(s, "device_name");
	sdr_context->last_timestamp = 0;
	sdr_context->frame_count = 0;
	sdr_context->send_mrs_status = ((uint64_t) 1 << TRX_RDMA_MR_COUNT) - 1;
	sdr_context->consumed_mr_id = -1;

	if(!sdr_context->device_name) {
		fprintf(stderr, "Missing `device_name` parameter\n");
		return -1;
	}

	s->opaque = sdr_context;
	s->trx_start_func2 = trx_rdma_start;
	s->trx_write_func2 = trx_rdma_write;
	s->trx_read_func2 = trx_rdma_read;
	s->trx_end_func = trx_rdma_end;

	return 0;
}

