#include "libtrx/trx_vcr.h"

int trx_vcr_start(TRXState *s, const TRXDriverParams2 *p) {
	struct TRXVCRContext *trx_vcr_ctx;

	size_t block_count;
	FILE *play_file_handle;
	uint32_t play_file_size;
	char *play_file_content;

	trx_vcr_ctx = s->opaque;

	// Play mode: allocate a buffer and fill it from file
	play_file_handle = fopen(trx_vcr_ctx->play_file_path, "r");
	if(!play_file_handle) {
		fprintf(stderr, "trx_vcr: file %s does not exist\n", trx_vcr_ctx->play_file_path);
		return -1;
	}

	// Allocate the right size
	fseek(play_file_handle, 0, SEEK_END);
	play_file_size = ftell(play_file_handle);
	rewind(play_file_handle);

	play_file_content = malloc(play_file_size);
	block_count = fread(play_file_content, play_file_size, 1, play_file_handle);

	if(block_count != 1) {
		fprintf(stderr, "trx_vcr: could not read samples from file\n");
		return -1;
	}

	// Use an infinite circular buffer with same length as the file
	trx_vcr_ctx->play_buffer = common_circular_buffer_create(play_file_size, true);
	common_circular_buffer_write(trx_vcr_ctx->play_buffer, (void *) play_file_content, play_file_size);

	printf("trx_vcr: looping %lu frames from %s\n", play_file_size / sizeof(TRXComplex), trx_vcr_ctx->play_file_path);

	fclose(play_file_handle);
	free(trx_vcr_ctx->play_file_path);

	// Record mode: allocate a buffer of given length
	if(trx_vcr_ctx->record_file_length && trx_vcr_ctx->record_file_path) {
		trx_vcr_ctx->record_buffer = malloc(trx_vcr_ctx->record_file_length * sizeof(TRXComplex));
	}

	return 0;
}

void trx_vcr_write(TRXState *s, trx_timestamp_t timestamp, const void **samples, int count, int tx_port_index, TRXWriteMetadata *md) {
	struct TRXVCRContext *trx_vcr_ctx;
	TRXComplex **channels;

	FILE *record_file_handle;

	trx_vcr_ctx = s->opaque;
	channels = (TRXComplex **) samples;

	// Early return when recording is not needed
	if(!trx_vcr_ctx->record_buffer) return;

	if(trx_vcr_ctx->record_file_length > trx_vcr_ctx->recorded_samples) {
		int sample_count = MIN(count, trx_vcr_ctx->record_file_length - trx_vcr_ctx->recorded_samples);

		// XXX: handle multiple channels
		memcpy(trx_vcr_ctx->record_buffer + trx_vcr_ctx->recorded_samples, (void *) channels[0], sample_count * sizeof(TRXComplex));
		trx_vcr_ctx->recorded_samples += sample_count;
	} else {
		// Write to file at the end to avoid disturbing sample collection
		record_file_handle = fopen(trx_vcr_ctx->record_file_path, "w");
		fwrite(trx_vcr_ctx->record_buffer, trx_vcr_ctx->record_file_length * sizeof(TRXComplex), 1, record_file_handle);
		fclose(record_file_handle);

		printf("trx_vcr: saving %u frames to %s\n", trx_vcr_ctx->record_file_length, trx_vcr_ctx->record_file_path);

		// Make sure to early return on next call to write
		trx_vcr_ctx->record_buffer = NULL;
	}
}

int trx_vcr_read(TRXState *s, trx_timestamp_t *ptimestamp, void **psamples, int count, int rx_port_index, TRXReadMetadata *md) {
	struct TRXVCRContext *trx_vcr_ctx;
	TRXComplex **channels;

	trx_vcr_ctx = s->opaque;
	channels = (TRXComplex **) psamples;

	// Ignoring result: reading an infinite buffer will always work
	common_circular_buffer_read(trx_vcr_ctx->play_buffer, (void *) channels[0], count * sizeof(TRXComplex));
	memcpy((void *) channels[1], (void *) channels[0], count * sizeof(TRXComplex));

	trx_vcr_ctx->last_timestamp += count;
	*ptimestamp = trx_vcr_ctx->last_timestamp;

	return count;
}

static void trx_vcr_end(TRXState *s) {
	struct TRXVCRContext *trx_vcr_ctx;

	trx_vcr_ctx = s->opaque;

	common_circular_buffer_free(trx_vcr_ctx->play_buffer);
	free(trx_vcr_ctx);
}

int trx_driver_init(TRXState *s) {
	struct TRXVCRContext *trx_vcr_ctx = malloc(sizeof(struct TRXVCRContext));

	char *record_file_length_string;

	if(s->trx_api_version != TRX_API_VERSION) {
		fprintf(stderr, "trx_vcr: ABI compatibility mismatch between LTEENB and TRX driver (LTEENB ABI version=%d, TRX driver ABI version=%d)\n", s->trx_api_version, TRX_API_VERSION);
		return -1;
	}

	record_file_length_string = trx_get_param_string(s, "record_length");
	if(record_file_length_string) {
		trx_vcr_ctx->record_file_length = atoi(record_file_length_string);
		free(record_file_length_string);
	}

	trx_vcr_ctx->record_buffer = NULL;
	trx_vcr_ctx->recorded_samples = 0;
	trx_vcr_ctx->record_file_path = trx_get_param_string(s, "record_file");

	trx_vcr_ctx->play_file_path = trx_get_param_string(s, "play_file");
	if(!trx_vcr_ctx->play_file_path) {
		fprintf(stderr, "trx_vcr: missing `play_file` parameter\n");
		return -1;
	}

	s->opaque = trx_vcr_ctx;
	s->trx_start_func2 = trx_vcr_start;
	s->trx_write_func2 = trx_vcr_write;
	s->trx_read_func2 = trx_vcr_read;
	s->trx_end_func = trx_vcr_end;

	return 0;
}

